# Realization of a multi-node quantum network of remote solid-state qubits
# Pompili, Hermans, Baier, et al.

import qutip as qt
import numpy as np

parameters = {
    "epr_ab": {
        "dc_rate": 10,
        "det_win": 15e-9,
        "pdet_1": 1.8e-4 * 2,
        "pdet_2": 2.2e-4 * 2,
        "vis": 0.9,
        "p2e": 0.06,
        "phase_u": 30,
        "lde_length": 3.8e-6,
        "depol": 0.0,
    },
    
    "epr_bc": {
        "dc_rate": 10,
        "det_win": 15e-9,
        "pdet_1": 2.1e-4 * 2,
        "pdet_2": 1.5e-4 * 2,
        "vis": 0.9,
        "p2e": 0.08,
        "phase_u": 15,
        "lde_length": 5.0e-6,
        "depol": 0.0,
    },
    
    "memory": {
        "depol":(0.11),
        "alphas":[0,0.05, 0.1, 0.15, 0.2, 0.3, 0.5],
        "N1e":[2116, 1843, 1697, 1383, 1202, 1386,1399],
        "power":[1.3, 1.4, 1.3, 1.3, 1.4, 1.3,1.1]
    },
    
    "decoupling": {  # BVL loss after decoupling
        "alice": 1.0-0.96,
        "charlie": 1.0-0.99,
    },
    
    "ionization_prob": {
        "alice": 0.015,
        "bob": 0.00,
        "charlie": 0.007
    },
    
    "bob_ro_fidelities": {
        "0": 0.930,
        "1": 0.994,
    }
}

def _actual_fidelity(x, y):
    return qt.fidelity(x, y)**2

def _circuit_list_to_propagator(circuit_list):
    propagator = circuit_list[-1]
    for g in reversed(circuit_list[:-1]):
        propagator *= g
    return propagator

def _transform_rho(rho, gate):
    return gate * rho * gate.dag()

I = qt.identity(2)
X = qt.sigmax()
Y = qt.sigmay()
Z = qt.sigmaz()
H = qt.operations.hadamard_transform(1)
II = qt.identity([2, 2])
ZI = qt.tensor(Z, I)
IZ = qt.tensor(I, Z)
ZZ = qt.tensor(Z, Z)
IH = qt.tensor(I, H)
IX = qt.tensor(I, X)
XI = qt.tensor(X, I)
IY = qt.tensor(I, Y)
YI = qt.tensor(Y, I)

zero = qt.basis(2,0)
one = qt.basis(2,1)
plus = (zero+one).unit()

X90 = qt.operations.rx(np.pi/2)
Y90 = qt.operations.ry(np.pi/2)
mX90 = qt.operations.rx(-np.pi/2)
mY90 = qt.operations.ry(-np.pi/2)
X180 = qt.operations.rx(np.pi)
Y180 = qt.operations.ry(np.pi)
CX90 = (
    qt.operations.controlled_gate(X90, control=1, target=0, control_value=0) 
    * qt.operations.controlled_gate(mX90, control=1, target=0, control_value=1))

ent_gate = _circuit_list_to_propagator([qt.tensor([I, mX90]), CX90, qt.tensor([I, X90])])
tensor_ent_gate = qt.tensor([I, ent_gate, I])

proj = qt.tensor(I,plus.proj(),I)

# GHZ generation

result_0 = qt.tensor([I, I, qt.projection(2, 0, 0), I])
result_1 = qt.tensor([I, I, qt.projection(2, 1, 1), I])

feedforward_0 = qt.tensor([I, I, I, mY90])
feedforward_1 = qt.tensor([I, I, I, Y90])
final_feedforward = qt.tensor([I, Y90, I, X180])

GHZ = (qt.tensor([qt.basis(2,0)]*3) + qt.tensor([qt.basis(2,1)]*3))/np.sqrt(2)

# Entanglement swapping

proj_0 = qt.projection(2, 0, 0)
proj_1 = qt.projection(2, 1, 1)

result_00 = qt.tensor([I, proj_0, proj_0, I])
result_01 = qt.tensor([I, proj_0, proj_1, I])
result_10 = qt.tensor([I, proj_1, proj_0, I])
result_11 = qt.tensor([I, proj_1, proj_1, I])

result_dict = {
    '00': result_00,
    '01': result_01,
    '10': result_10,
    '11': result_11,
}

feedforward_00 = qt.tensor([I, I, I, X180*mY90])
feedforward_01 = qt.tensor([I, I, I, X180*Y90])
feedforward_10 = qt.tensor([I, I, I, Y180*mY90])
feedforward_11 = qt.tensor([I, I, I, Y180*Y90])

feedforward_dict = {
    '00': feedforward_00,
    '01': feedforward_01,
    '10': feedforward_10,
    '11': feedforward_11,
}

bell_ideal = qt.ket2dm(qt.bell_state('00'))

def _dephase(rho, p):
    return (1-p)*rho + IZ*rho*IZ * p

def _depolarise(rho, p, target=1):
    if target==0:
        return (1-p)*rho + (ZI*rho*ZI + XI*rho*XI + YI*rho*YI) * p/3.
    elif target==1:
        return (1-p)*rho + (IZ*rho*IZ + IX*rho*IX + IY*rho*IY) * p/3.
    else:
        assert False, "target either 0 or 1"
        
def msmt_operator(op, node):
    # POVMs for Z readout including ionization
    F_0 = qt.Qobj(np.array([[1-parameters["ionization_prob"][node], 0], [0, 0]]), dims=[[2], [2]])
    F_1 = qt.Qobj(np.array([[parameters["ionization_prob"][node], 0], [0, 1.0]]), dims=[[2], [2]])
    
    toX = qt.operations.rotation(Y, -np.pi/2)
    toY = qt.operations.rotation(X, np.pi/2)
    
    if op=="I":
        return I
    elif op=="Z":
        return F_0 - F_1
    elif op=="X":
        return toX.dag() * (F_0 - F_1) * toX
    elif op=="Y":
        return toY.dag() * (F_0 - F_1) * toY

def sce_state(alpha_1, alpha_2, pars, psi_plus=True):
    assert 0 < alpha_1 < 1, f"alpha_1 = {alpha_1} not in range"
    assert 0 < alpha_2 < 1, f"alpha_2 = {alpha_2} not in range"
    
    Pdc = pars["dc_rate"] * pars["det_win"]   # Dark count prob in window
    
    Puu = alpha_1 * alpha_2 * (
        (1 - Pdc)**2 * (pars["pdet_1"] * (1 - pars["pdet_2"]) + pars["pdet_2"] * (1 - pars["pdet_1"])) 
        + 2*(1 - Pdc) * Pdc * (1 - pars["pdet_1"]) * (1 - pars["pdet_2"])
    )
    Pud = alpha_1 * (1 - alpha_2) * ((1 - Pdc)**2 * pars["pdet_1"] + 2*Pdc * (1 - Pdc) * (1 - pars["pdet_1"]))
    Pdu = (1 - alpha_1) * alpha_2 * ((1 - Pdc)**2 * pars["pdet_2"] + 2*Pdc * (1 - Pdc) * (1 - pars["pdet_2"]))
    Pdd = 2 * (1 - alpha_1) * (1 - alpha_2) * Pdc * (1 - Pdc)
    
    Pe = 0.5 * (1 - np.exp(-0.5 * (pars["phase_u"]/180*np.pi)**2)) # Phase uncertainty
    
    Ptot = Puu + Pud + Pdu + Pdd
    rate = Ptot / pars["lde_length"]
    fidelity = (1-pars["depol"]) * (Pud + Pdu + 2 * (1 - pars["p2e"])**2 * (1 - 2*Pe) * np.sqrt(pars["vis"] * Pud * Pdu)) / (2*Ptot) + pars["depol"]/4
    
    rho = qt.Qobj(np.array([
            [Puu, 0, 0, 0],
            [0, Pud, np.sqrt(pars["vis"] * Pud * Pdu), 0],
            [0, np.sqrt(pars["vis"] * Pud * Pdu), Pdu, 0],
            [0, 0, 0, Pdd]]) / Ptot, dims=[[2, 2], [2, 2]])

    # Simulate double excitation
    rho = (1 - pars["p2e"]/2) ** 2 * rho \
          + (1 - pars["p2e"]/2) * (pars["p2e"]/2) * ZI * rho * ZI \
          + (1 - pars["p2e"]/2) * (pars["p2e"]/2) * IZ * rho * IZ \
          + (pars["p2e"]/2) ** 2 * ZZ * rho * ZZ
    
    # Phase uncertainty
    rho = (1 - Pe) * rho + Pe * IZ * rho * IZ

    # Depolarizing noise      
    rho = (1-pars["depol"])*rho + pars["depol"] * II/4
    
    if not psi_plus:
        rho = _transform_rho(rho, IZ)
    
    return rho, rate, fidelity

def ghz_state(alpha_1_a, alpha_1_b, alpha_2_b, alpha_2_c, timeout, RO_result="all"):
    
    assert timeout >= 1, "Wrong timeout"
    timeout = int(timeout)
    
    rho_ab, rate_ab, _ = sce_state(alpha_1_a, alpha_1_b, parameters["epr_ab"])
    
    # Depolarize Alice's qubit because of dynamical decoupling
    rho_ab = _depolarise(rho_ab, parameters["decoupling"]["alice"] * 0.75, target=0)
        
    # Swap operation mapping is equivalent to applying Hadamard
    rho_ab = _transform_rho(rho_ab, IH)
    
    # Depolarise and dephase memory
    rho_ab = _depolarise(rho_ab, parameters["memory"]["depol"] * 0.75)
    N1e = np.interp(alpha_2_b, parameters["memory"]["alphas"], parameters["memory"]["N1e"])
    power = np.interp(alpha_2_b, parameters["memory"]["alphas"], parameters["memory"]["power"])

    #We average over possible memory dephasings.
    p_deph = (1 - np.exp(-(np.arange(1, timeout) / N1e)**power)).mean()
    rho_ab = _dephase(rho_ab, p_deph * 0.5)
    
    rho_bc, rate_bc, _ = sce_state(alpha_2_b, alpha_2_c, parameters["epr_bc"])
    
    # Depolarize Charlie's qubit because of dynamical decoupling
    rho_bc = _depolarise(rho_bc, parameters["decoupling"]["charlie"] * 0.75, target=1)
        
    # 0 - communication alice, 1 - memory bob, 2 - communication bob, 3 - communication charlie
    full_rho = qt.tensor([rho_ab, rho_bc])
    
    # Apply entangling gate
    full_rho = _transform_rho(full_rho, tensor_ent_gate)
    
    F0 = parameters["bob_ro_fidelities"]["0"]
    F1 = parameters["bob_ro_fidelities"]["1"]
    
    final_rho = 0
    if (RO_result in ["0", "all"]):
        final_rho += F0 * _transform_rho(full_rho, _circuit_list_to_propagator([result_0, feedforward_0]))
        final_rho += (1 - F1) * _transform_rho(full_rho, _circuit_list_to_propagator([result_1, feedforward_0]))
    if (RO_result in ["1", "all"]):
        final_rho += (1 - F0) * _transform_rho(full_rho, _circuit_list_to_propagator([result_0, feedforward_1])) 
        final_rho += F1 * _transform_rho(full_rho, _circuit_list_to_propagator([result_1, feedforward_1]))
        
    final_rho = _transform_rho(final_rho, final_feedforward)
    ghz_rho = final_rho.ptrace([0, 1, 3]).unit()
    
    p_succ_before_timeout = 1 - (1 - rate_bc * parameters["epr_bc"]["lde_length"])**timeout
    rate_ghz = p_succ_before_timeout / (1/rate_ab + 1/rate_bc)
    fid_ghz = _actual_fidelity(ghz_rho, GHZ)
    
    return ghz_rho, rate_ghz, fid_ghz

ghz_operators_list = ["XXX", "XYY", "YXY", "YYX", "IZZ", "ZIZ", "ZZI"]

ghz_operators = {
    op: qt.tensor([msmt_operator(m_o, n) for m_o, n in zip(op, ["alice", "bob", "charlie"])]) for op in ghz_operators_list
}

def swapped_state(alpha_1_a, alpha_1_b, alpha_2_b, alpha_2_c, timeout, BSM_result="all"):
    
    assert timeout >= 1, "Wrong timeout"
    timeout = int(timeout)
    
    rho_ab, rate_ab, _ = sce_state(alpha_1_a, alpha_1_b, parameters["epr_ab"])
    
    # Depolarize Alice's qubit because of dynamical decoupling
    rho_ab = _depolarise(rho_ab, parameters["decoupling"]["alice"] * 0.75, target=0)
        
    # Swap operation mapping is equivalent to applying Hadamard
    rho_ab = _transform_rho(rho_ab, IH)
    
    # Depolarise and dephase the memory
    rho_ab = _depolarise(rho_ab, parameters["memory"]["depol"] * 0.75)
    N1e = np.interp(alpha_2_b, parameters["memory"]["alphas"], parameters["memory"]["N1e"])
    power = np.interp(alpha_2_b, parameters["memory"]["alphas"], parameters["memory"]["power"])
    
    # We average over possible memory dephasings.
    p_deph = (1 - np.exp(-(np.arange(1, timeout) / N1e)**power)).mean()
    rho_ab = _dephase(rho_ab, p_deph * 0.5)
    
    rho_bc, rate_bc, _ = sce_state(alpha_2_b, alpha_2_c, parameters["epr_bc"])
    
    # Depolarize Charlie's qubit because of dynamical decoupling
    rho_bc = _depolarise(rho_bc, parameters["decoupling"]["charlie"] * 0.75, target=1)
        
    # 0 - communication alice, 1 - memory bob, 2 - communication bob, 3 - communication charlie
    full_rho = qt.tensor([rho_ab, rho_bc])
    
    # Apply entangling gate
    full_rho = _transform_rho(full_rho, tensor_ent_gate)
    
    ro_fids = parameters["bob_ro_fidelities"]
    wrong_ro = {
        "0": "1",
        "1": "0"
    }
    
    final_rho = 0
    
    for r_c in ["0", "1"]:
        for r_e in ["0", "1"]:
            result_proj = result_dict[r_c+r_e]
            # Both correct
            if ((BSM_result=="all") or (r_c+r_e == BSM_result)):
                final_rho += ro_fids[r_c]*ro_fids[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedforward_dict[r_c+r_e]]))
                
            # r_c correct, r_e wrong
            if ((BSM_result=="all") or (r_c+wrong_ro[r_e] == BSM_result)):
                final_rho += ro_fids[r_c]*(1-ro_fids[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedforward_dict[r_c+wrong_ro[r_e]]]))
                
            # r_c wrong, r_e correct
            if ((BSM_result=="all") or (wrong_ro[r_c]+r_e == BSM_result)):
                final_rho += (1-ro_fids[r_c])*ro_fids[r_e]*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedforward_dict[wrong_ro[r_c]+r_e]]))
                
            # both wrong
            if ((BSM_result=="all") or (wrong_ro[r_c]+wrong_ro[r_e] == BSM_result)):
                final_rho += (1-ro_fids[r_c])*(1-ro_fids[r_e])*_transform_rho(full_rho, _circuit_list_to_propagator([result_proj, feedforward_dict[wrong_ro[r_c]+wrong_ro[r_e]]]))
    
    swapped_rho = final_rho.ptrace([0, 3]).unit()
    
    p_succ_before_timeout = 1 - (1 - rate_bc * parameters["epr_bc"]["lde_length"])**timeout
    rate_ghz = p_succ_before_timeout / (1/(rate_ab*1) + 1/(rate_bc*1))
    fid_swapped = _actual_fidelity(swapped_rho, bell_ideal)
    
    return swapped_rho, rate_ghz, fid_swapped

swapped_operators_list = ["XX", "YY", "ZZ"]

swapped_operators = {
    op: qt.tensor([msmt_operator(m_o, n) for m_o, n in zip(op, ["alice", "charlie"])]) for op in swapped_operators_list
}

def get_2node_correlators(pair, alpha_1, alpha_2, psi_plus=True):

    XtoZ = [qt.tensor(mY90,I),qt.tensor(I,mY90)]
    YtoZ = [qt.tensor(X90,I), qt.tensor(I,X90)]

    expectations = {}

    for key in swapped_operators.keys():
        if "I" in key: continue
        if "E" in key: continue
        rho_bell = bell_ideal
        rho_sce = sce_state(alpha_1, alpha_2, parameters[pair], psi_plus)[0]
        for j in range(2):
            if key[j] == 'X':
                rho_sce = _transform_rho(rho_sce, XtoZ[j])
            elif key[j] == 'Y':
                rho_sce = _transform_rho(rho_sce, YtoZ[j])

        G = np.real_if_close(rho_sce.diag())
        outcomes = ['00','01','10','11']
        expectations[key] = {outcomes[i]: G[i] for i in range(len(outcomes))}
    return expectations